;;; ediff-minor.el -- use ediff commands from your ediff buffers

;;;   Author: Greg Klanderman <greg.klanderman@alum.mit.edu>
;;;  Version: 0.0
;;;     Date: June 13, 1998
;;;  Created: June 8, 1998
;;; Keywords: diff patch emacs ediff

;;; Copyright (C) 1998 Greg Klanderman
;;;
;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 2, or (at your option)
;;; any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; A copy of the GNU General Public License can be obtained from the
;;; Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;;; Boston, MA 02111-1307, USA.
;;;
;;; Send bug reports, feedback, etc. to greg.klanderman@alum.mit.edu, 
;;; greg@alphatech.com, or gregk@ai.mit.edu.

;;; COMMENTARY:
;;;
;;; This is a package for emacs which allows you to operate ediff directly
;;; from the buffers being compared rather than from the ediff control
;;; buffer.  Putting the commands
;;;
;;; (add-hook 'ediff-load-hook
;;;           (lambda ()
;;;             (setq ediff-grab-mouse nil)
;;;             (require 'ediff-minor)))
;;;
;;; in your .emacs should allow you to use ediff commands usually
;;; only available from the ediff control buffer from any of your
;;; ediff buffers.  The command C-c C-c allows you to switch between
;;; ediff-minor mode and ediff-edit-minor mode, the latter being a
;;; no-op other than providing C-c C-c to get back to ediff-minor
;;; mode.
;;;
;;; #### OK, that's a lie, at least for now.  Use
;;;
;;; (setq ediff-grab-mouse nil)
;;; (require 'ediff-minor)
;;;
;;; until I can get it working via the hook.

;;; TODO:
;;;
;;;  * Test more than next/previous difference, quit,
;;;    copy-A-to-B, restore-diff.
;;;  * Lots of error checking needed!!!
;;;  * Fix enabling via ediff-load-hook.
;;;  * Don't auto-insinuate.
;;;  * Handle the case of multiple active ediff sessions
;;;    involving the same buffer.  I have a solution, it
;;;    just needs to get implemented once the rest of this
;;;    stabilizes some.

;;; HISTORY:
;;;
;;; Jun 10, 1998: Version 0.0
;;;  * First released to xemacs-beta.

;;; BRAINDAMAGE:
;;;
;;; There seems to be no way to have a minor-mode-map be buffer
;;; local.  Ideally we'd like to put a symbol pointing to the
;;; keymap ediff-minor-mode-map in the minor-mode-map-alist, but
;;; that results in looking in the function slots, which can not be
;;; buffer local.  So we resort to an ugly hack where
;;; ediff-minor-mode variants are created on the fly as needed.
;;; There are only 24 different possible mode maps, after all.

;;; ======================================================================

;;; Code:

(require 'ediff)


;;;
;;; buffer local variables
;;;

;; store pointer to the control buffer
(defvar ediff-minor-control-buffer nil)
(make-variable-buffer-local 'ediff-minor-control-buffer)

;(defvar ediff-minor-mode nil)
;(make-variable-buffer-local 'ediff-minor-mode)

;; store symbol of actual ediff-minor-mode variant
(defvar ediff-minor-mode-variant nil)
(make-variable-buffer-local 'ediff-minor-mode-variant)

;(defvar ediff-minor-mode-map nil)
;(make-variable-buffer-local 'ediff-minor-mode-map)

;; ediff-edit-minor-mode
(defvar ediff-edit-minor-mode nil)
(make-variable-buffer-local 'ediff-edit-minor-mode)


;;;
;;; modeline strings
;;;

(defvar ediff-minor-mode-string      " Ediff-minor")
(defvar ediff-edit-minor-mode-string " Ediff-edit")


;;;
;;; ediff-minor startup hook
;;;

(defun ediff-minor-startup ()
  (let* ((control (current-buffer))
         (varid   (ediff-minor-variant-id))
         (variant (intern (concat "ediff-minor-mode:" varid)))
         (mapname (intern (concat "ediff-minor-map:"  varid))))
    (when (or (not (boundp mapname))
              (not (assq variant minor-mode-map-alist)))
      (let ((keymap (ediff-minor-setup-keymap mapname)))
        (set mapname keymap)
        (add-minor-mode variant 'ediff-minor-mode-string keymap)))
    (mapc (lambda (buf)
            (when buf
              (save-excursion
                (set-buffer buf)
                (ediff-minor-mode t variant control))))
          (list ediff-buffer-A ediff-buffer-B ediff-buffer-C))))

(defun ediff-minor-variant-id ()
  (format "%s-%s-%s-%s"
          (if ediff-word-mode  "wm" "**")
          (if ediff-narrow-job "nj" "**")
          (if ediff-3way-job   "3w" "**")
          (cond (ediff-merge-job "mj")
                (ediff-3way-comparison-job "3w")
                (t "**"))))

(add-hook 'ediff-startup-hook 'ediff-minor-startup)


;;;
;;; ediff-minor quit, suspend hook
;;;

(defun ediff-minor-shutdown ()
  (mapc (lambda (buf)
          (when buf
            (save-excursion
              (set-buffer buf)
              (ediff-minor-mode nil))))
        (list ediff-buffer-A ediff-buffer-B ediff-buffer-C)))

(add-hook 'ediff-suspend-hook 'ediff-minor-shutdown)
(add-hook 'ediff-quit-hook    'ediff-minor-shutdown)


;;;
;;; ediff-minor-mode
;;;

(defun ediff-minor-mode (turnon &optional variant control)
  (cond (turnon
         (setq ediff-minor-mode-variant   variant
               ediff-minor-control-buffer control)
         (set (make-local-variable variant) t))
        (t
         (set  ediff-minor-mode-variant   nil)
         (setq ediff-minor-mode-variant   nil
               ediff-minor-control-buffer nil)))
  (redraw-modeline))

; blech... this does not work.
;
;(add-minor-mode 'ediff-minor-mode
;                'ediff-minor-mode-string
;                'ediff-minor-mode-map)


;;;
;;; ediff-edit-minor-mode
;;;

(defun ediff-minor-toggle-edit ()
  "Toggle between ediff-minor-mode and ediff-edit-minor-mode."
  (interactive)
  (let ((minorp (symbol-value-in-buffer ediff-minor-mode-variant
                                        (current-buffer))))
    (set  ediff-minor-mode-variant (not minorp))
    (setq ediff-edit-minor-mode    minorp)
    (redraw-modeline)))

(defvar ediff-edit-minor-map nil
  "Keymap for ediff-edit-minor-mode.")

(unless ediff-edit-minor-map
  (setq ediff-edit-minor-map
        (make-sparse-keymap 'ediff-edit-minor-map))
  (define-key ediff-edit-minor-map "\C-c\C-c" 'ediff-minor-toggle-edit))

(add-minor-mode 'ediff-edit-minor-mode
                'ediff-edit-minor-mode-string
                ediff-edit-minor-map)


;;;
;;; build ediff-minor-mode keymap
;;;

(defun ediff-minor-setup-keymap (mapname)
  (let ((map (make-sparse-keymap mapname)))
    (suppress-keymap map)

;    (define-key map
;      (if ediff-emacs-p [mouse-2] [button2]) 'ediff-help-for-quick-help)
;    (define-key map "\C-m"  'ediff-help-for-quick-help)
    (define-key map "p" 'ediff-minor-previous-difference)
    (define-key map "\C-?" 'ediff-minor-previous-difference)
    (define-key map [delete] 'ediff-minor-previous-difference)
;    (define-key map "\C-h" (if ediff-no-emacs-help-in-control-buffer
;                               'ediff-minor-previous-difference nil))
    ;; must come after C-h, or else C-h wipes out backspace's binding in XEmacs
    (define-key map [backspace] 'ediff-minor-previous-difference)
    (define-key map "n" 'ediff-minor-next-difference)
    (define-key map " " 'ediff-minor-next-difference)
    (define-key map "j" 'ediff-minor-jump-to-difference)
    (define-key map "g"  nil)
    (define-key map "ga" 'ediff-minor-jump-to-difference-at-point)
    (define-key map "gb" 'ediff-minor-jump-to-difference-at-point)
    (define-key map "q" 'ediff-minor-quit)
    (define-key map "D" 'ediff-minor-show-diff-output)
    (define-key map "z" 'ediff-minor-suspend)
    (define-key map "\C-l" 'ediff-minor-recenter)
    (define-key map "|" 'ediff-minor-toggle-split)
    (define-key map "h" 'ediff-minor-toggle-hilit)
    (or ediff-word-mode
        (define-key map "@" 'ediff-minor-toggle-autorefine))
    (if ediff-narrow-job
        (define-key map "%" 'ediff-minor-toggle-narrow-region))
    (define-key map "~" 'ediff-minor-swap-buffers)
    (define-key map "v" 'ediff-minor-scroll-vertically)
    (define-key map "\C-v" 'ediff-minor-scroll-vertically)
    (define-key map "^" 'ediff-minor-scroll-vertically)
    (define-key map "\M-v" 'ediff-minor-scroll-vertically)
    (define-key map "V" 'ediff-minor-scroll-vertically)
    (define-key map "<" 'ediff-minor-scroll-horizontally)
    (define-key map ">" 'ediff-minor-scroll-horizontally)
    (define-key map "i" 'ediff-minor-status-info)
    (define-key map "E" 'ediff-minor-documentation)
    (define-key map "?" 'ediff-minor-toggle-help)
    (define-key map "!" 'ediff-minor-update-diffs)
    (define-key map "M" 'ediff-minor-show-current-session-meta-buffer)
    (define-key map "R" 'ediff-minor-show-registry)
    (or ediff-word-mode
        (define-key map "*" 'ediff-minor-make-or-kill-fine-diffs))
    (define-key map "a"  nil)
    (define-key map "b"  nil)
    (define-key map "r"  nil)
    (cond (ediff-merge-job
           ;; Will barf if no ancestor
           (define-key map "/" 'ediff-minor-show-ancestor)
           ;; In merging, we allow only A->C and B->C copying.
           (define-key map "a" 'ediff-minor-copy-A-to-C)
           (define-key map "b" 'ediff-minor-copy-B-to-C)
           (define-key map "r" 'ediff-minor-restore-diff-in-merge-buffer)
           (define-key map "s" 'ediff-minor-shrink-window-C)
           (define-key map "+" 'ediff-minor-combine-diffs)
           (define-key map "$" 'ediff-minor-toggle-show-clashes-only)
           (define-key map "&" 'ediff-minor-re-merge))
          (ediff-3way-comparison-job
           (define-key map "ab" 'ediff-minor-copy-A-to-B)
           (define-key map "ba" 'ediff-minor-copy-B-to-A)
           (define-key map "ac" 'ediff-minor-copy-A-to-C)
           (define-key map "bc" 'ediff-minor-copy-B-to-C)
           (define-key map "c" nil)
           (define-key map "ca" 'ediff-minor-copy-C-to-A)
           (define-key map "cb" 'ediff-minor-copy-C-to-B)
           (define-key map "ra" 'ediff-minor-restore-diff)
           (define-key map "rb" 'ediff-minor-restore-diff)
           (define-key map "rc" 'ediff-minor-restore-diff)
           (define-key map "C"  'ediff-minor-toggle-read-only))
          (t                                 ; 2-way comparison
           (define-key map "a"  'ediff-minor-copy-A-to-B)
           (define-key map "b"  'ediff-minor-copy-B-to-A)
           (define-key map "ra" 'ediff-minor-restore-diff)
           (define-key map "rb" 'ediff-minor-restore-diff))
          )                                  ; cond
    (define-key map "G" 'ediff-minor-submit-report)
    (define-key map "#"  nil)
    (define-key map "#h"  'ediff-minor-toggle-regexp-match)
    (define-key map "#f"  'ediff-minor-toggle-regexp-match)
    (or ediff-word-mode
        (define-key map "##"  'ediff-minor-toggle-skip-similar))
    (define-key map "o"   nil)
    (define-key map "A"  'ediff-minor-toggle-read-only)
    (define-key map "B"  'ediff-minor-toggle-read-only)
    (define-key map "w"   nil)
    (define-key map "wa"  'ediff-minor-save-buffer)
    (define-key map "wb"  'ediff-minor-save-buffer)
    (define-key map "wd"  'ediff-minor-save-buffer)
    (define-key map "="   'ediff-minor-inferior-compare-regions)
;    (if (fboundp 'ediff-minor-show-patch-diagnostics)
;        (define-key map "P"  'ediff-minor-show-patch-diagnostics))
    (if ediff-3way-job
        (progn
          (define-key map "wc" 'ediff-minor-save-buffer)
          (define-key map "gc" 'ediff-minor-jump-to-difference-at-point)
          ))

    (define-key ediff-mode-map "m" 'ediff-minor-toggle-wide-display)
	
    ;; the one addition with respect to ediff-mode-map:
    ;;
    (define-key map "\C-c\C-c" 'ediff-minor-toggle-edit)

    map))


;;;
;;; with-selected-window for ediff
;;;

(defmacro ediff-with-selected-window (window &rest body)
  "Execute forms in BODY with WINDOW as the selected window.
The value returned is the value of the last form in BODY."
  (let ((old-win (gensym "ediff-wsw")))
    `(let ((,old-win (selected-window)))
       (unwind-protect
           (progn (select-window ,window t)
                  ,@body)
         (select-window ,old-win t)))))
(put 'ediff-with-selected-window 'lisp-indent-function 1)


;;;
;;; macros to wrap ediff commands -> ediff-minor commands
;;;

(defmacro ediff-minor-wrap-body (&rest body)
  `(let ((ctl-buff ediff-minor-control-buffer))
     (if (and (symbol-value-in-buffer
               ediff-minor-mode-variant (current-buffer))
              ctl-buff
              (buffer-live-p ctl-buff))
         (let ((ctl-wind (save-excursion
                           (set-buffer ctl-buff)
                           ediff-control-window)))
           (if (window-live-p ctl-wind)
               (ediff-with-selected-window ctl-wind
                 ,@body)
             (error "bad ediff-control-window")))
       (error "not working from an ediff-minor-mode buffer"))))

(defun ediff-minor-make-arglist (lo hi)
  (let ((args nil))
    (while (< lo hi)
      (setq lo (1+ lo))
      (setq args
            (cons (intern (concat "arg-" (int-to-string lo))) args)))
    (nreverse args)))

(defmacro ediff-minor-wrap (name)
  (let* ((funcname  (intern (concat "ediff-" (symbol-name name))))
         (func      (symbol-function funcname))
         (max-arg   (function-max-args func))
         (min-arg   (function-min-args func))
         (std-args  (ediff-minor-make-arglist 0 min-arg))
         (opt-args  (ediff-minor-make-arglist min-arg max-arg))
         (wrap-args (append std-args (if opt-args
                                         (append '(&optional) opt-args))))
         (call-args (append std-args opt-args))
         (interact  (function-interactive func))
         (wrapped   (intern (concat "ediff-minor-" (symbol-name name)))))
    `(defun ,wrapped ,wrap-args
       ,interact
       (ediff-minor-wrap-body
        (,funcname ,@call-args)))))


;;;
;;; wrap various ediff commands
;;;

(ediff-minor-wrap previous-difference)
(ediff-minor-wrap next-difference)
(ediff-minor-wrap jump-to-difference)
(ediff-minor-wrap jump-to-difference-at-point)
(ediff-minor-wrap quit)
(ediff-minor-wrap show-diff-output)
(ediff-minor-wrap suspend)
(ediff-minor-wrap recenter)
(ediff-minor-wrap toggle-split)
(ediff-minor-wrap toggle-hilit)
(ediff-minor-wrap toggle-autorefine)
(ediff-minor-wrap toggle-narrow-region)
(ediff-minor-wrap swap-buffers)
(ediff-minor-wrap scroll-vertically)
(ediff-minor-wrap scroll-horizontally)
(ediff-minor-wrap status-info)
(ediff-minor-wrap documentation)
(ediff-minor-wrap toggle-help)
(ediff-minor-wrap update-diffs)
(ediff-minor-wrap show-current-session-meta-buffer)
(ediff-minor-wrap show-registry)
(ediff-minor-wrap make-or-kill-fine-diffs)
(ediff-minor-wrap show-ancestor)
(ediff-minor-wrap copy-A-to-C)
(ediff-minor-wrap copy-B-to-C)
(ediff-minor-wrap restore-diff-in-merge-buffer)
(ediff-minor-wrap shrink-window-C)
(ediff-minor-wrap combine-diffs)
(ediff-minor-wrap toggle-show-clashes-only)
(ediff-minor-wrap re-merge)
(ediff-minor-wrap copy-A-to-B)
(ediff-minor-wrap copy-B-to-A)
(ediff-minor-wrap copy-C-to-A)
(ediff-minor-wrap copy-C-to-B)
(ediff-minor-wrap restore-diff)
(ediff-minor-wrap toggle-read-only)
(ediff-minor-wrap submit-report)
(ediff-minor-wrap toggle-regexp-match)
(ediff-minor-wrap toggle-skip-similar)
(ediff-minor-wrap save-buffer)
(ediff-minor-wrap inferior-compare-regions)
(ediff-minor-wrap toggle-wide-display)


;;;
;;; provide feature
;;;

(provide 'ediff-minor)


;;; ediff-minor.el ends here
